program PiBAS;
uses
  BScript;

const
  DefaultInput  = 'PROGRAM.PBS';
  DefaultOutput = 'RESULT.LST';

  EmptyErrorCode = -1;

procedure Error(Message: String; ErrorCode: Integer);
begin
  Writeln('Internal program error!');
  if ErrorCode <> EmptyErrorCode
  then Writeln(Message, ErrorCode)
  else Writeln(Message);
  Halt(255);
end;

var
  S: BigString;
  Index: Integer;
  InputFileName: String;
  OutputFileName: String;
  OutFile: Text;
  InFile: Text;
  IOResultTemp: Integer;
  IsIntrospective: Boolean;
  CurChar: Char;

begin
  if ((ParamCount = 1) and (ParamStr(1) = '/?')) or (ParamCount > 2)
  then begin
    if ParamCount > 2
    then Writeln('Illegal parameters.');
    Writeln;
    Writeln('PiBAS interpreter. (C) Rybinsk, 2002 by Kopachev Michael.');
    Writeln('  To use:');
    Writeln('    PIBAS.EXE /? | /c | [InputFile [ResultFile]] ');
    Writeln('      InputFile  - PiBAS program file (default - ', DefaultInput,')');
    Writeln('      OutputFile - file of PiBAS program execution result');
    Writeln('                   (defalut - ', DefaultOutput, ')');
    Writeln('      /c         - Console mode (InputFile = OutputFile = ''CON'')');
    Writeln('      /?         - This screen');
    Writeln;
  end else begin
    InputFileName := DefaultInput;
    OutputFileName := DefaultOutput;
    if ParamCount = 1
    then begin
      if (ParamStr(1) = '/c') or (ParamStr(1) = '/C')
      then begin
        InputFileName := '';
        OutputFileName := '';
      end else InputFileName := ParamStr(1);
    end else if ParamCount = 2
             then begin
               InputFileName := ParamStr(1);
               OutputFileName := ParamStr(2);
             end;
    {Try open input file}
    Assign(OutFile,InputFileName);
    {$I-}
    Reset(OutFile);
    IOResultTemp := IOResult;
    {$I+}
    if IOResultTemp <> 0
    then Error('Can''t open "' + InputFileName + '" file to read. DOS error code #', IOResultTemp)
    else Close(OutFile);
    Interprete(InputFileName, S);
    Assign(OutFile, OutputFileName);
    {$I-}
    Rewrite(OutFile);
    IOResultTemp := IOResult;
    {$I+}
    if IOResultTemp <> 0
    then Error('Can''t open "' + OutputFileName + '" file to write. DOS error code #', IOResultTemp);
    IsIntrospective := True;
    Assign(InFile, InputFileName);
    Reset(InFile);
    for Index := 1 to S^.StringSize do begin
      Write(OutFile, S^.Body[Index]);
      Read(InFile, CurChar);
      IsIntrospective := IsIntrospective and (CurChar = S^.Body[Index]);
    end;
    while eoln(InFile) and (not eof(InFile)) do
      Readln(InFile);
    IsIntrospective := IsIntrospective and eof(InFile);
    Writeln;
    if IsIntrospective
    then Writeln('This program is introspective! Length = ', S^.StringSize);
    Finalize(S);
    Close(OutFile);
    Close(InFile);
  end;
end.