{ ᮤন ᠭ ⮪,  ਬ⨢ ࠡ  }
unit Streams;

interface

const
  LastID = 1;
  ErrorOccur: boolean = False;
  LastErrorMessage: string = '';

type
  TIDEntry = record
    Name: string[20];
    Value: Longint;
  end;
  TIDs = array [1..LastID] of TIDEntry;
  TStreamSet = set of byte;

  TLastSymbol = (symNone, symNumber, symString, symSymbol, symID, symEndStream);
  TAbstractStream = object
  private
    FCurrentState: TLastSymbol;
    FNumber: Longint;
    FString: string;
    FSymbol: Char;
    FID: String;
    procedure   OpenStream; virtual;
    procedure   CloseStream; virtual;
    procedure   ReadLexeme; virtual; {Main procedure ந ࠧ 室!}
  public
    constructor Create;
    destructor  Destroy;
    function    ReadNumber: Longint;
    function    ReadSymbol: Char;
    function    ReadString: string;
    function    ReadID: String;
    procedure   ReadSet(var Result);
    procedure   StartBlock;
    procedure   EndBlock;
    procedure   NextTeg;
    procedure   WaitEqual;
    procedure   WaitSymbol(Symbol: char);
    function    GetNextLexeme: TLastSymbol;
    function    CurrentLexeme: TLastSymbol;
    function    ReadChar: Char;
    procedure   Error(Message: string); virtual;
  end;

  PTextFileStream = ^TTextFileStream;

  TErrorProcedure = procedure (const Message: String; Stream: PTextFileStream);


  TTextFileStream = object (TAbstractStream)
  private
    FFile: Text;
    FFileName: string;
    FIsFirst: Boolean;
    CurrentChar: Char;
    QuoteChar: Char;
    procedure   OpenStream; virtual;
    procedure   CloseStream; virtual;
    procedure   ReadLexeme; virtual; {Main procedure ந ࠧ 室!}
  public
    CurrentLine: Longint;
    CurrentSymbol: Longint;
    ErrorLine: Longint;
    ErrorSymbol: Longint;
    Lexeme: String;
    ErrorHandler: TErrorProcedure;
    constructor Create(const FileName: string);
    procedure   Error(Message: string); virtual;
  end;

  TMemoryStream = object (TAbstractStream)
    Current: Pointer;
    procedure   ReadLexeme; virtual; {Main procedure ந ࠧ 室!}
  public
    constructor Create(Start: Pointer);
  end;

const

  IDs: TIDs = ((Name: ''; Value: -1));


implementation

{TAbstractStream}

constructor TAbstractStream.Create;
begin
  OpenStream;
  FCurrentState := symNone;
end;

destructor TAbstractStream.Destroy;
begin
  CloseStream;
end;

procedure TAbstractStream.OpenStream;
begin
end;

procedure TAbstractStream.CloseStream;
begin
end;

procedure TAbstractStream.ReadLexeme;
begin
end;

procedure TAbstractStream.Error;
begin
  Writeln(Message);
  ErrorOccur := True;
  LastErrorMessage := Message;
end;

function TAbstractStream.ReadNumber;
begin
  while (FCurrentState <> symNumber) and (FCurrentState <> symEndStream) do
    ReadLexeme;
  if FCurrentState <> symNumber
  then Error('Unexpected end of stream!')
  else ReadNumber := FNumber;
end;

function TAbstractStream.ReadSymbol;
begin
  while (FCurrentState <> symSymbol) and (FCurrentState <> symEndStream) do
    ReadLexeme;
  if FCurrentState <> symSymbol
  then Error('Unexpected end of stream!')
  else ReadSymbol := FSymbol;
end;

function TAbstractStream.ReadString;
begin
  while (FCurrentState <> symString) and (FCurrentState <> symEndStream) do
    ReadLexeme;
  if FCurrentState <> symString
  then Error('Unexpected end of stream!')
  else ReadString := FString;
end;

function TAbstractStream.ReadID;
begin
  while (FCurrentState <> symID) and (FCurrentState <> symEndStream) do
    ReadLexeme;
  if FCurrentState <> symID
  then Error('Unexpected end of stream!')
  else ReadID := FID;
end;

procedure TAbstractStream.StartBlock;
begin
  WaitSymbol('{');
end;

procedure TAbstractStream.EndBlock;
begin
  WaitSymbol('}');
end;

procedure TAbstractStream.NextTeg;
begin
  repeat
    while (FCurrentState <> symSymbol) and (FCurrentState <> symEndStream) do
      ReadLexeme;
    if FCurrentState <> symSymbol
    then Error('Unexpected end of stream!');
  until FSymbol in [';', '}'];
end;

procedure TAbstractStream.WaitEqual;
begin
  WaitSymbol('=');
end;

procedure TAbstractStream.WaitSymbol;
begin
  repeat
    if FSymbol <> Symbol
    then ReadLexeme;
    while (FCurrentState <> symSymbol) and (FCurrentState <> symEndStream) do
      ReadLexeme;
    if FCurrentState = symEndStream
    then Error('Unexpected end of stream!');
  until FSymbol = Symbol;
end;

function TAbstractStream.GetNextLexeme;
begin
  ReadLexeme;
  GetNextLexeme := FCurrentState;
end;

function TAbstractStream.CurrentLexeme;
begin
  CurrentLexeme := FCurrentState;
end;

procedure TAbstractStream.ReadSet;
var
  CurSym: Char;
  Quit: boolean;
  Next: boolean;
begin
  WaitSymbol('[');
  Quit := False;
  Next := True;
  repeat
    case GetNextLexeme of
      symNumber: begin
                   Include(TStreamSet(Result), Byte(ReadNumber));
                   Next := True;
                 end;
      symSymbol: begin
                   case ReadSymbol of
                     ',': if Next
                          then Next := False
                          else Error('Incorrect set format!');
                     ']': if Next
                          then Quit := True
                          else Error('Incorrect set format!');
                     #0..#32: ;
                   else Error('Incorrect set qualifier!');
                   end;
                 end;
    else Error('Incorrect set format!');
    end;
  until Quit;
end;

function TAbstractStream.ReadChar;
var
  TempString: string;
begin
  TempString := ReadString;
  if Length(TempString) <> 1
  then ReadChar := TempString[1]
  else Error('Icorrect char defenition');
end;

{TTextFileStream}

constructor TTextFileStream.Create;
begin
  FFileName := FileName;
  Inherited Create;
  CurrentChar := ' ';
  CurrentLine := 1;
  CurrentSymbol := 0;
  ErrorLine := 1;
  ErrorSymbol := 0;
  ErrorHandler := Nil;
end;

procedure TTextFileStream.Error;

function IntToStr(Value: Longint): String;
var
  Result: String;
begin
  Str(Value, Result);
  IntToStr := Result;
end;

begin
  if Assigned(ErrorHandler)
  then ErrorHandler(Message, @Self)
  else begin
    Inherited Error(Message);
    Writeln('At line ', CurrentLine, ', position ', CurrentSymbol);
    Halt(255);
  end;
end;

procedure  TTextFileStream.OpenStream;
var
  F: File of Char;

begin
  ErrorLine := 0;
  ErrorSymbol := 0;
  FIsFirst := True;
  Lexeme := '';
  Assign(F, FFileName);
  {$I-}
  Reset(F);
  {$I+}
  if IOResult <> 0
  then Error('Error open read ' + FFileName + '.')
  else begin
    if FileSize(F) > 32000
    then begin
      Close(F);
      Error('Input file too long.');
    end else Close(F);
  end;
  Assign(FFile, FFileName);
  {$I-}
  Reset(FFile);
  {$I+}
  if IOResult <> 0
  then Error('Error open read ' + FFileName + '.');
end;

procedure TTextFileStream.CloseStream;
begin
  {$I-}
  Close(FFile);
  {$I+}
end;

procedure  TTextFileStream.ReadLexeme;
type
  TAutoState = (asStart, asString, asNumber, asID, asStringLast);

var
  State: TAutoState;
  Quit: boolean;
  IDName: string;
  NumNegative: boolean;
  Index: integer;
  Found: boolean;
begin
  State := asStart;
  Quit := False;
  Lexeme := '';
  repeat
    if CurrentChar in [#0..#32]
    then begin
      if EOF(FFile)
      then begin
        if FIsFirst
        then Error('Unexpected end of file.');
        {Check correct states}
        if State = asStart
        then begin
          FCurrentState := symEndStream;
          Exit;
        end else begin
          if State = asString
          then Error('Unexpected end of file.')
          else CurrentChar := #0;
        end;
      end else begin
        if EOLN(FFile)
        then Error('Unexpected CR/LF.')
        else begin
          Read(FFile, CurrentChar);
          Inc(CurrentSymbol);
        end;
      end;
    end;
    FIsFirst := FIsFirst and (CurrentChar in [#0..#32]);
    case State of
      asStart: begin
                 Lexeme := CurrentChar;
                 {砫쭮 ﭨ -   ⮣,  ⠥ }
                 CurrentChar := UpCase(CurrentChar);
                 case CurrentChar of
                   '_',
                   'A'..'Z': begin
                               State := asID;
                               IDName := CurrentChar;
                               CurrentChar := ' ';
                               ErrorLine := CurrentLine;
                               ErrorSymbol := CurrentSymbol;
                             end;
                   '"', '''': begin
                     State := asString;
                     FString := '';
                     QuoteChar := CurrentChar;
                     CurrentChar := ' ';
                     ErrorLine := CurrentLine;
                     ErrorSymbol := CurrentSymbol;
                   end; {'"', ''''}
                   '0'..'9': begin
                               State := asNumber;
                               if CurrentChar in ['1'..'9']
                               then FNumber := Byte(CurrentChar) - Byte('0')
                               else FNumber := 0;
                               CurrentChar := ' ';
                               NumNegative := CurrentChar = '-';
                               ErrorLine := CurrentLine;
                               ErrorSymbol := CurrentSymbol;
                             end;
                   #0..#32: {஡ ᨬ -  ய᪠};
                   else begin
                     {}
                     FSymbol := CurrentChar;
                     CurrentChar := ' ';
                     Quit := True;
                     FCurrentState := symSymbol;
                     ErrorLine := CurrentLine;
                     ErrorSymbol := CurrentSymbol;
                   end;
                 end;
               end;
      asString: begin
                  {७ ப}
                  if CurrentChar <> QuoteChar
                  then FString := FString + CurrentChar
                  else begin
                    Quit := True;
                    FCurrentState := symString;
                  end;
                  Lexeme := Lexeme + CurrentChar;
                  CurrentChar := ' ';
                end;
      asNumber: begin
                  {᫮}
                  if CurrentChar in ['0'..'9']
                  then begin
                    FNumber := FNumber * 10 + Byte(CurrentChar) - Byte('0');
                    Lexeme := Lexeme + CurrentChar;
                    CurrentChar := ' ';
                  end else begin
                    Quit := True;
                    FCurrentState := symNumber;
                    if NumNegative
                    then FNumber := -FNumber;
                  end;
                end;
      asID: begin
              if UpCase(CurrentChar) in ['0'..'9', 'A'..'Z', '_']
              then begin
                IDName := IDName + UpCase(CurrentChar);
                Lexeme := Lexeme + CurrentChar;
                CurrentChar := ' ';
              end else begin
                {饬 䨪,  ⠭  祭}
                Quit := True;
                FCurrentState := symID;
                FID := IDName;
              end;
            end;
      else Error('Internal error');
    end;
  until Quit;
end;

{TMemory Stream}

constructor TMemoryStream.Create;
begin
  Current := Start;
  Inherited Create;
end;

procedure TMemoryStream.ReadLexeme;
type
  THeader = record
    State: TLastSymbol;
    Size: Word;
    Data: Longint;
  end;
var
  Header: THeader;
begin
  FCurrentState := THeader(Current^).State;
  case FCurrentState of
    symNumber: FNumber := THeader(Current^).Data;
    symString: Move(THeader(Current^).Data, FString, Byte(THeader(Current^).Data) + 1);
    symSymbol: FSymbol := Char(THeader(Current^).Data);
  end;
  Inc(Longint(Current), THeader(Current^).Size);
end;

end.